/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mouse Gestures for Mozilla.
 *
 * The Initial Developer of the Original Code is Pavol Vaskovic.
 * Portions created by the Initial Developer are Copyright (C) 2001
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s) (alphabetical order):
 *  Andy Edmonds <aedmonds@mindspring.com>
 *  Benjamin K. Stuhl <tiriath@yahoo.com>
 *  David Illsley <illsleydc@bigfoot.com>
 *  Dorando <mozilla@dorando.at>
 *  HJ van Rantwijk <bugs4HJ@netscape.net>
 *  Jens Bannmann <jens.b@web.de>
 *  Jochen <autohide@krickelkrackel.de>
 *  Martin.T.Kutschker <Martin.T.Kutschker@blackbox.net>
 *  Pavol Vaskovic <pali@pali.sk>
 *  Pekka Aleksi Knuutila <aleksi.knuutila@edu.lahti.fi>
 *  Scott R. Turner <srt@aero.org>
 *  Tim Williamson <chsman@hotmail.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// Global variables
// mouse gesture condition (to be evaluated on event) for more details see mozgestReadPref()
var mgContext=null;
var mgContent=null;

var mgButton = -1;
var condition = new String;
var grid; // grid size; minimal gesture has to be 'grid' pixels long
var delay; // delay before aborting gesture
var lmbGestureLimit;
var dragdropDelay;
var gestureStartTime;
var mgLastGestureTime; // end timestamp of the last gesture attempt
var rockersEnabled;
var mgWheelRockers;
var mgWheelSensitivity;
var dTolerance; // tolerance for diagonal movement. see processCoordinates()
var gestureInProgress = false;
var mgRockerWithRMB = false;
var mgPreviousSelection = null;
// old coordinates from previous gesture stroke used in processCoordinates()
var old_x;
var old_y;
var gridMoves = new Array();
var gesture = new Array();
var localizedGesture = new Array();
var mgString; // array holding localized stroke names
var lastgesture;
// lastGestureTime is time of last gesture stroke, compared with delay and current time
var lastGestureTime;
var lastMoveTime;
// timer used to abort gestures after delay, see processCoordinates()
var timer;

// trail settings
var drawTrail;
var trailLimit;
var trailWidth;
var trailInterval;
var trailColor;
var trailOpacity;

// globalOnLink is an array that holds a list of all the links traversed during gesture
// globalOnImage is a string containing an image href or false
// globalSrcEvent stores event which started the gesture the active gesture.
// The links are used in the gesture functions in gesture.js.
var globalOnLink = false;
var globalOnImage = false;
var globalSrcEvent = false;
var bundle = null; // String bundle for localized strings
var mgAllowContextMenu = false;
var mgLastHoveredElement = null;

// used by some gesture implementations
const kWindowMediator = Components
      .classes["@mozilla.org/appshell/window-mediator;1"]
      .getService(Components.interfaces.nsIWindowMediator);

// preferences root branch, used by gesture implementations
var prefRoot = Components.classes["@mozilla.org/preferences-service;1"]
                 .getService(Components.interfaces.nsIPrefService)
                 .getBranch("");

// Preferences observer object; implements nsIObserver
function mgPrefObserver()
{
  try {
    var pbi = pref.QueryInterface(Components.interfaces.nsIPrefBranchInternal);
    pbi.addObserver(this.domain, this, false);
  } catch(ex) {
    dump("MozGest: Failed to observe prefs: " + ex + "\n");
  }
}
mgPrefObserver.prototype = {
  domain: "mozgest.",
  observe: function(subject, topic, prefName) {
    // when MozGest pref was changed, we reinitialize
    if (prefName != "mozgest.sideBarSort") mozgestInit();
  }
}

// This code is performed on load:
// get the preferences service
var prefService = Components.classes["@mozilla.org/preferences-service;1"]
                            .getService(Components.interfaces.nsIPrefService);
var pref = prefService.getBranch(null); // preferences root node
var mgPref = prefService.getBranch("mozgest."); // preferences mozgest node
window.mozgestPrefObserver = new mgPrefObserver(); // and set an Pref observer
// END of on load performed code

function mgContextMenuListener()
{   removeEventListener("popupshowing", mgDisableContextMenu, true)
    addEventListener("popupshowing", mgDisableContextMenu, true)
    setTimeout("removeEventListener('popupshowing', mgDisableContextMenu, true)",0)
}

function mgDisableContextMenu(e)
{
    mgContext=e.originalTarget
    removeEventListener("popupshowing", mgDisableContextMenu, true)
    if (!mgAllowContextMenu)
       e.preventDefault();
}

function mgAllowContextByKeyPress(e)
{   if ((e.keyCode==121 && e.shiftKey) || e.keyCode==93)
       mgAllowContextMenu=true
}

function mgStartup(e) {
  // we *should* remove our handler here, but that may cause a bug.
  // see http://bugzilla.mozilla.org/show_bug.cgi?id=174320
  // instead, we stay registered and just check whether this is the correct
  // onload event.
  if (e.view != window) {
    return;
  }
  // window.removeEventListener("load", mgStartup, true);

  determineWindowType();
  if (windowType) {
    dump("MozGest: Startup\n");
    dockElement("stringbundleset", "stringbundle", "mozgestbundle",
        new Array("src"),
        new Array("chrome://mozgest/locale/mozgest.properties"));
    dockElement("keyset", "key", "mgPrefKey",
        new Array("modifiers",   "key", "command"  ),
        new Array("accel shift", "M",   "mgPrefCmd"));
    dockElement("commandset", "command", "mgPrefCmd",
        new Array("oncommand"),
        new Array("showMgPrefPanel();"));

    window.addEventListener("unload", mgStorage.saveGestureCounts, true);
    mozgestLoadStrings();
    mozgestInit();
  }
}

function mozgestInit() {
  // read prefs (or set default for each missing setting)
  mozgestReadPref();

  // turn gestures recognition on/off
  if (mgIsEnabled()) {
    addWindowWatch();
    loadGestureTable();
  } else {
    removeWindowWatch();
  }
}

function loadGestureTable(aBool) {
  dump("Loading gesture table\n");
  resetGesturesTable();
  mgStorage.initUserRDF();
  mgStorage.readRDFGestureTable("window");
  mgStorage.readRDFGestureTable(windowType);

  // notify the sidebar(s), do this only once if there is more than one window
  if (aBool && kWindowMediator.getMostRecentWindow("navigator:browser") == window) {
     Components.classes["@mozilla.org/observer-service;1"]
               .getService(Components.interfaces.nsIObserverService)
               .notifyObservers(null, "mgMappingsUpdated", null);
  }
}

function mozgestLoadStrings() {
  // load localized string bundle and cache often used strings
  bundle = document.getElementById("mozgestbundle");
  mgString = new Array();
  mgString["R"] = bundle.getString("abbreviation.right");
  mgString["L"] = bundle.getString("abbreviation.left");
  mgString["U"] = bundle.getString("abbreviation.up");
  mgString["D"] = bundle.getString("abbreviation.down");
  mgString["1"] = bundle.getString("abbreviation.d1");
  mgString["3"] = bundle.getString("abbreviation.d3");
  mgString["7"] = bundle.getString("abbreviation.d7");
  mgString["9"] = bundle.getString("abbreviation.d9");
  mgString["gesture"] = bundle.getString("g.gesture");
}

function processCoordinates(e){ // e is browser's event object
  var g = gesture.join("");
  if (condition == "e.button==0" && (g == "L" || g == "R")
      && (e.timeStamp - gestureStartTime) > lmbGestureLimit) {
    endGesture(null);
  }

  mozgestDrawTrail(e);

  var x_dir = e.clientX - old_x;
  var y_dir = e.clientY - old_y;
  var x = Math.abs(x_dir);
  var y = Math.abs(y_dir);

  if ( (x >= grid/2 || y >= grid/2) ) { // process each half-grid
    // diagonal movement:
    // dTolerance = 75 means that a movement is recognized as diagonal when
    // x/y or y/x is between 0.25 and 1
    if ( dTolerance != 0 && ( (x/y >= (1-dTolerance/100) && x/y <= 1)
        || (y/x >= (1-dTolerance/100) && y/x <= 1) ) ) {

      if (x_dir < 0 && y_dir > 0) {
        gPush("1");
      } else if (x_dir > 0 && y_dir > 0) {
        gPush("3");
      } else if (x_dir < 0 && y_dir < 0) {
        gPush("7");
      } else if (x_dir > 0 && y_dir < 0) {
        gPush("9");
      }

    }
    // horizontal movement:
    else if (x > y) {
      if (x_dir > 0) {
        gPush("R");
      } else if (x_dir < 0) {
        gPush("L");
      }
    }
    // vertical movement:
    else if (x < y) {
      if (y_dir > 0){
        gPush("D");
      } else if (y_dir < 0){
        gPush("U");
      }
    }
    mgMsg(mgString["gesture"] + " " + localizedGesture);
    lastGestureTime = e.timeStamp;
    if (timer) { window.clearTimeout(timer); }
    timer = window.setTimeout("endGesture(null);", delay);

    old_x = e.clientX;
    old_y = e.clientY;
  }
  lastMoveTime = e.timeStamp;

  mgFindHoveredElements(e.originalTarget);
}

function gPush(code) {
  gridMoves.push(code);
  if (gridMoves.length == 2) {
    if (gridMoves[0] == gridMoves[1] || gesture.length == 0) {
      doPush(gridMoves.pop());
    }
    gridMoves.length = 0;
  }
}

function doPush(code) {
  if (gesture[gesture.length-1] != code) {
    gesture.push(code);
    localizedGesture.push(mgString[code]);
  }
  if (gesture.length == 1) resetRocker();
}

function mgFindHoveredElements(domPosition) {
  var distance = 0;
  while (domPosition != null
         && domPosition != mgContent
         && distance < 5) {
    mgExamineHoveredElement(domPosition);
    domPosition = domPosition.parentNode;
    distance++;
  }
}

function mgExamineHoveredElement(node) {
  var tag = node.nodeName.toLowerCase();
  if (tag == "a" && node.href && !mozgestCheckLinkDups(node.href)) {
    if(!globalOnLink) globalOnLink = new Array();
    globalOnLink.push(node);
    dump("Mozgest: hovered link '" + node.href + "'\n");
  } else if (tag == "img" && globalOnImage == false) {
    globalOnImage = node;
    dump("Mozgest: found image.\n");
  }
}

function mozgestCheckLinkDups(addy) { // avoid duplicates in globalOnLink array
  var isADup = false;
  for(var zz = 0; zz < globalOnLink.length; zz++) {
    if(globalOnLink[zz].href == addy) isADup = true;
  }
  return isADup;
}

// startGesture is called from event listener in bubble phase
function startGesture(e){
  mgAllowContextMenu = false;

  // collect ancestor elements, which are needed by inContent checks
  var ancestors = new Array();
  var cur = e.originalTarget;
  while (cur != null && cur != mgContent) {
    var tag = cur.nodeName.toLowerCase();
    if (!ancestors[tag]) ancestors[tag] = cur;
    cur = cur.parentNode;
  }

  if (mozgestIsInContent(ancestors))
  {
    globalOnLink.length = 0;
    globalOnImage = false;
    if(!checkForRocker(e) // note: rockers *can* start while gestureInProgress
       && !gestureInProgress
       && eval(condition)) {
      globalSrcEvent = e;
      gestureInProgress = true;
      gestureStartTime = e.timeStamp;
      lastMoveTime = 0;
      old_x = e.clientX;
      old_y = e.clientY;

      var sel = globalSrcEvent.view.getSelection();
      if (mgButton == 0 && sel.rangeCount > 0)
      {
        mgPreviousSelection = sel.getRangeAt(0);
      }

      // check if we started over link or image
      if (ancestors["a"]) mgExamineHoveredElement(ancestors["a"]);
      if (ancestors["img"]) mgExamineHoveredElement(ancestors["img"]);

      gesture.length = 0; // clear gesture array
      localizedGesture.length = 0; // clear localized gesture array
      mozgestBeginTrail(e);
      window.addEventListener("mousemove", processCoordinates, false);
      lastGestureTime = e.timeStamp;
    }
  } else {
    mgRockerWithRMB=false;
  }
}

function endGesture(e) {
  releaseRocker(e);
  var noGestureDone = false;
  if (gestureInProgress) {
    window.removeEventListener("mousemove", processCoordinates, false);
    mozgestEndTrail();
    window.clearTimeout(timer);
    lastgesture = gesture.join("");

    // check if we've been called via a timeout; see processCoordinates()
    if (e == null) {
      dump("gesture timeouted\n");
      mgMsg(bundle.getString("g.aborted"));
    } else if (e == "drag") {
    } else if (lastgesture != ""){
      window.setTimeout(fireGesture, 1, lastgesture);

      // we need to save the current timestamp here, so if a click event
      // follows, it can be detected as related to this mouseup event
      // (see mgOnClickHandler)
      mgLastGestureTime = (new Date).getTime();
    } else {
      dump("MozGest: Gesture was too small\n");
      noGestureDone = true;
    }

    gestureInProgress = false;
    gestureStartTime = false;
  } else {
    noGestureDone = true;
  }

  if (e && (e.button == 2 || (navigator.platform == "MacPPC"
           && e.button == 0 && e.ctrlKey)))
  {
     if (!mgRockerWithRMB && noGestureDone)
     {
        if (navigator.platform == "Win32" || navigator.platform == "OS/2")
           mgAllowContextMenu = true;

        else                              // linux & mac
        { document.popupNode=e.originalTarget;
          if (e.screenY >= mgContent.boxObject.screenY &&
              e.screenY <= mgContent.boxObject.screenY + mgContent.boxObject.height &&
              e.screenX >= mgContent.boxObject.screenX &&
              e.screenX <= mgContent.boxObject.screenX +mgContent.boxObject.width)

                  mgContext.showPopup(mgContext.ownerDocument.documentElement, e.clientX, e.clientY,
                                          "context", "bottomleft", "topleft");
        }  // end no win32/os2
     }
  }
}

function mozgestReadPref(){ // read settings from Preferences interface
  grid = mgGetPref("grid", "Int", 15, 10, 500);
  delay = mgGetPref("delay", "Int", 1000, 500, 10000);
  lmbGestureLimit = mgGetPref("lmbGestureLimit", "Int", 500, 100, 10000);
  dragdropDelay = mgGetPref("dragdropDelay", "Int", 400, 100, 10000);
  rockersEnabled = mgGetPref("enableRockers", "Bool", true);
  mgWheelRockers = mgGetPref("wheelRockers", "Bool", false);
  mgWheelSensitivity = mgGetPref("wheelSensitivity", "Int", 1, 1, 9);
  dTolerance = mgGetPref("diagonalTolerance", "Int", 60, 0, 75);

  // condition is a boolean expression consisting of event object values
  mgButton = mgGetPref("mousebutton", "Int", 0, 0, 2);
  condition = "e.button==" + mgButton;
  if (mgGetPref("modifier.ctrl", "Bool", false))
    condition += " && e.ctrlKey";
  if (mgGetPref("modifier.alt", "Bool", false))
    condition += " && e.altKey";
  if (mgGetPref("modifier.shift", "Bool", false))
    condition += " && e.shiftKey";

  drawTrail = mgGetPref("trails.enabled", "Bool", false);
  trailWidth = mgGetPref("trails.width", "Int", 5, 1, 100);
  trailInterval = mgGetPref("trails.interval", "Int", 2, 1, 100);
  trailColor = mgGetPref("trails.color", "Char", "#FF0000");
  trailLimit = mgGetPref("trails.length-limit", "Int", 1000, 1, 10000);
  trailOpacity = mgGetPref("trails.opacity", "Int", 100, 1, 100);
  mgSideBarSort = mgGetPref("sideBarSort", "Char", "sortByUsage");
}

// Gets a pref setting. If it's missing, set it to a default value.
// For Int prefs, you can optionally specify a minimum and maximum value.
function mgGetPref(name, type, defaultValue, min, max) {
  if ((type != "Bool" && type != "Char" && type != "Int") || !name) {
    return null;
  }
  var notfound = false;
  var value;
  try {
    value = eval("mgPref.get" + type + "Pref(\"" + name + "\");");
  } catch (e) {
    notfound = true;
  }
  if (notfound || (type == "Int" && (min || max) && (value > max || value < min))) {
    val = (type == "Char") ? ("\"" + defaultValue + "\"") : defaultValue;
    eval("mgPref.set" + type + "Pref(\"" + name + "\", " + val + ");");
    return defaultValue;
  } else {
    return value;
  }
}

function addWindowWatch() {
  dump("MozGest: ---- Adding Listeners ---\n");
  if (document.documentElement.getAttribute("windowtype")=="navigator:browser")
     mgContent = mgGetContentArea().mPanelContainer;
  else
     mgContent = mgGetContentArea();

  mgContent.addEventListener("mousedown", startGesture, false);
  mgContent.addEventListener("draggesture", onDragGestureEvent, true);
  if (mgWheelRockers) {
    mgContent.addEventListener("DOMMouseScroll", mgMousewheelHandler, false);
  }
  window.addEventListener("mouseup", endGesture, false);
  window.addEventListener("keydown", mgAllowContextByKeyPress, true);
  mgContent.addEventListener("contextmenu", mgContextMenuListener, true);
  mgContent.addEventListener("click", mgOnClickHandler, true);
}

function removeWindowWatch() {
  dump("MozGest: ---- Removing Listeners ---\n");
  mgContent.removeEventListener("mousedown", startGesture, false);
  mgContent.removeEventListener("draggesture", onDragGestureEvent, true);
  mgContent.removeEventListener("DOMMouseScroll", mgMousewheelHandler, false)
  mgContent.removeEventListener("contextmenu", mgContextMenuListener, true)
  window.removeEventListener("mouseup", endGesture, false);
  window.removeEventListener("keydown", mgAllowContextByKeyPress, true);
  mgContent.removeEventListener("click", mgOnClickHandler, true);
}

function mgOnClickHandler(e) {
  var now = (new Date).getTime();
  if (e.button == mgButton && now - mgLastGestureTime <= 1) {
    dump("MozGest: Stopping 'click' event caused by last gesture.\n");
    e.preventDefault();
    e.stopPropagation();
    return false;
  }

  var node = e.originalTarget;
  if (!node.getAttribute) return null;
  if (!node.hasAttribute("href")) return null;
  var tagname = node.nodeName.toLowerCase();
  var href = node.getAttribute("href");
  if (tagname == "a" && href.indexOf("mozgest://") == 0) {
    dump("MozGest: Link click " + href + "\n");
    mgStorage.processURL(href);
    e.stopPropagation();
    e.preventDefault();
  }
  return null;
}

function onDragGestureEvent(e) {
  var draggedElement = e.originalTarget.nodeName.toUpperCase();
  if (gestureInProgress && draggedElement != "HTML") {
    if (lastMoveTime-gestureStartTime > dragdropDelay) {
      dump("MozGest: '" + draggedElement
           + "' drag encountered, canceling gesture\n");
      endGesture("drag");
    } else {
      dump("MozGest: canceling '" + draggedElement
           + "' drag event, continuing gesture\n");
      e.stopPropagation();
    }
  }
}

function showMgPrefPanel() {
  window.openDialog("chrome://mozgest/content/pref/mozgestPrefWindow.xul",
                    "mozgestprefs", "chrome,centerscreen,modal");
}

function dockElement(setType, type, id, aNames, aValues) {
  dockElementAt(setType, type, id, aNames, aValues, document.documentElement);
}

function dockElementAt(setType, type, id, aNames, aValues, container) {
  var sets = document.getElementsByTagName(setType);
  var set;
  if (sets && sets.length > 0) {
    set = sets[0];
  } else {
    set = document.createElement(setType);
    container.appendChild(set);
  }
  var element = document.createElement(type);
  element.setAttribute("id", id);
  for (i=0; i < aNames.length; i++) {
    element.setAttribute(aNames[i], aValues[i]);
  }
  set.appendChild(element);
}